import 'context.dart';
import 'filters.dart';
import 'loaders.dart';
import 'nodes.dart';
import 'parser.dart';
import 'tests.dart';

/// The core component of Jinja 2 is the [Environment]. It contains
/// important shared variables like configuration, filters, tests and others.
/// Instances of this class may be modified if they are not shared and if no
/// template was loaded so far.
///
/// Modifications on environments after the first template was loaded
/// will lead to surprising effects and undefined behavior.
class Environment {
  /// Environment constructor. Load all templates from [loader] to cache.
  Environment({
    this.stmtOpen = '{%',
    this.stmtClose = '%}',
    this.varOpen = '{{',
    this.varClose = '}}',
    this.commentOpen = '{#',
    this.commentClose = '#}',
    this.defaultValue,
    this.loader,
    this.autoReload = true,
    Map<String, Function> filters = const <String, Function>{},
    Map<String, Function> tests = const <String, Function>{},
  })  : _templates = <String, Template>{},
        filters = Map<String, Function>.of(FILTERS)..addAll(filters),
        tests = Map<String, Function>.of(TESTS)..addAll(tests) {
    if (loader != null) loader.load(this);
  }

  /// The string marking the beginning of a statement.
  final String stmtOpen;

  /// The string marking the end of a statement.
  final String stmtClose;

  /// The string marking the beginning of a variable statement.
  final String varOpen;

  /// The string marking the end of a variable statement.
  final String varClose;

  /// The string marking the beginning of a comment.
  final String commentOpen;

  /// The string marking the end of a comment.
  final String commentClose;

  /// Map of custom Jinja filters to use.
  final Map<String, Function> filters;

  /// Map of custom Jinja tests to use.
  final Map<String, Function> tests;

  /// Default value for undefined values.
  final dynamic defaultValue;

  /// The template loader for this environment.
  final Loader loader;

  /// Some loaders load templates from locations where the template
  /// sources may change (ie: file system or database).  If
  /// [autoReload] is set to `true` (default) every time
  /// a template source changed loader reload template.
  /// For higher performance it's possible to disable that.
  final bool autoReload;

  final Map<String, Template> _templates;

  /// Load a template from a string. This parses the source given and
  /// returns a [Template] object.
  Template fromSource(String source, {String path, bool store = false}) {
    Template template = Template(this, source, path: path);
    if (store) if (path != null)
      _templates[path] = template;
    else
      throw ArgumentError.notNull('path');
    return template;
  }

  Template getTemplate(String path) {
    if (!_templates.containsKey(path))
      throw Exception('Template not found: $path');
    return _templates[path];
  }

  Function getFilter(String name) {
    if (!filters.containsKey(name)) throw Exception('Filter not found: $name');
    return filters[name];
  }

  Function getTest(String name) {
    if (!tests.containsKey(name)) throw Exception('Test not found: $name');
    return tests[name];
  }

  @override
  String toString() => 'Env($autoReload, $loader)';
}

/// The central template object. This class represents a compiled template
/// and is used to evaluate it.
class Template {
  Template(this.env, String source, {String path})
      : path = path,
        body = Parser(env).parse(source, path: path);

  final Environment env;
  final String path;
  final Node body;

  /// If no arguments are given the context will be empty.
  /// This will return the rendered template as unicode string.
  String render([Map<String, dynamic> context]) =>
      body.render(Context(env, data: context ?? const <String, dynamic>{}));

  /// If no arguments are given the context will be empty.
  /// This will return the rendered template as unicode string.
  String renderContext(Context context) => body.render(context);

  @override
  bool operator ==(dynamic other) =>
      other is Template && other.path == path && other.body == body;

  @override
  int get hashCode => path.hashCode ^ body.hashCode ^ 290418677;

  @override
  String toString() => 'Template($body${path != null ? ', ' + path : ''})';
}
